/**
 * 位置
 */
export class Location {
	#line: number;
	#column: number;
	#offset: number;

	constructor(line: number, column: number, offset: number) {
		this.#line = line;
		this.#column = column;
		this.#offset = offset;
	}

	/** 行号 */
	public get line(): number { return this.#line; }

	/** 列号 */
	public get column(): number { return this.#column; }

	/** 偏移 */
	public get offset(): number { return this.#offset; }
}

/** 位置信息 */
export class Position {
	#start: Location;
	#end: Location;

	constructor(start: Location, end: Location) {
		this.#start = start;
		this.#end = end;
	}

	/** 起始位置 */
	public get start(): Location { return this.#start; }

	/** 结束位置 */
	public get end(): Location { return this.#end; }


	/** 长度 */
	public get length(): number { return this.#end.offset - this.#start.offset; }
}

/** 节点 */
export abstract class Node {
	#position: Position;

	constructor(position: Position) {
		this.#position = position;
	}

	/** 位置信息 */
	public get position(): Position { return this.#position; }
}

/** 属性名 */
export class ElementAttributeName extends Node {
	#name: string;

	constructor(position: Position, name: string) {
		super(position);
		this.#name = name;
	}

	/** 属性名 */
	public get name(): string { return this.#name; }
}

/** 属性值 */
export class ElementAttributeValue extends Node {
	#value: string;

	constructor(position: Position, value: string) {
		super(position);
		this.#value = value;
	}

	/** 属性值 */
	public get value(): string { return this.#value; }
}

/** 元素标签 */
export class ElementTag extends Node {
	#tagName: string;

	constructor(position: Position, tagName: string) {
		super(position);
		this.#tagName = tagName;
	}

	/** 标签名 */
	public get tagName(): string { return this.#tagName; }
}

/** 元素属性 */
export class ElementAttribute extends Node {
	#name: ElementAttributeName;
	#value: ElementAttributeValue | null;
	#control: boolean;

	constructor(name: ElementAttributeName, value: ElementAttributeValue | null, control: boolean) {
		super(new Position(name.position.start, (value ?? name).position.end));
		this.#name = name;
		this.#value = value;
		this.#control = control;
	}

	/** 属性名 */
	public get name(): ElementAttributeName { return this.#name; }

	/** 属性值 */
	public get value(): ElementAttributeValue | null { return this.#value; }

	/** 是否控制属性 */
	public get control(): boolean { return this.#control; }
}

/** 元素 */
export class Element extends Node {
	#tag: ElementTag;
	#attributes: ElementAttribute[];
	#children: (Element | Text | Comment)[];
	#selfClosing: boolean;

	constructor(position: Position, tag: ElementTag, attributes: ElementAttribute[], selfClosing: boolean, children: (Element | Text | Comment)[]) {
		super(position);
		this.#tag = tag;
		this.#attributes = attributes;
		this.#selfClosing = selfClosing;
		this.#children = children;
	}

	/** 标签 */
	public get tag() { return this.#tag; }

	/** 属性列表 */
	public get attributes() { return this.#attributes; }

	/** 标签名 */
	public get tagName() { return this.#tag.tagName; }

	/** 是否自闭合 */
	public get selfClosing() { return this.#selfClosing; }

	/** 子节点列表 */
	public get children() { return this.#children; }
}

/** 文本节点 */
export class Text extends Node {
	#value: string;

	constructor(position: Position, value: string) {
		super(position);
		this.#value = value;
	}

	/** 文本内容 */
	public get value(): string { return this.#value; }
}

/** 文档类型 */
export class DocType extends Node {
	#items: string[];

	constructor(position: Position, items: string[]) {
		super(position);
		this.#items = items;
	}

	/** 文档类型信息 */
	public get items(): string[] { return this.#items; }
}

/** 注释 */
export class Comment extends Node {
	#value: string;

	constructor(position: Position, value: string) {
		super(position);
		this.#value = value;
	}

	/** 注释内容 */
	public get value(): string { return this.#value; }
}

/** 文档 */
export class Document {
	#doctype: DocType | null = null;
	#children: (Element | Text | Comment)[] = [];

	/** 文档类型 */
	public set doctype(doctype: DocType | null) { this.#doctype = doctype; }
	public get doctype() { return this.#doctype; }

	/** 子节点列表 */
	public get children() { return this.#children; }

	/** 添加子节点 */
	public addChild(child: Element | Text | Comment) { this.#children.push(child); }
}
