import fs from 'fs';
import { Comment, DocType, Document, Element, ElementAttribute, ElementAttributeName, ElementAttributeValue, ElementTag, Location, Position, Text } from './elem';
import { ParseError } from './error';

export class Parser {
	#filePath: string;
	#content: string;
	#line: number = 0;
	#column: number = 0;
	#offset: number = 0;
	#document!: Document;

	private constructor(filePath: string, content: string) {
		this.#filePath = filePath;
		this.#content = content;
	}

	public static fromFile(filePath: string) { return new Parser(filePath, fs.readFileSync(filePath, 'utf-8')); }

	public static fromString(content: string, filePath: string) { return new Parser(filePath, content); }

	public parse() {
		this.#line = 0;
		this.#column = 0;
		this.#offset = 0;
		this.#document = new Document();
		this.#parse().forEach(e => this.#document.addChild(e));
		return this.#document;
	}

	private is(text: string, offset: number = 0, caseSensitive: boolean = false) {
		const sub = this.#content.slice(this.#offset + offset, this.#offset + offset + text.length);
		if (caseSensitive) return sub === text;
		else return sub.toLowerCase() === text.toLowerCase();
	}

	private go(length: number) {
		for (let i = 0; i < length; i++) {
			if (this.#content[this.#offset + i] === '\n') {
				this.#line++;
				this.#column = 0;
			}
			else this.#column++;
		}
		this.#offset += length;
	}

	private eof() {
		return this.#offset >= this.#content.length;
	}

	private eat(text: string, caseSensitive: boolean = false) {
		if (this.is(text, 0, caseSensitive)) {
			this.go(text.length);
			return true;
		}
		return false;
	}

	private eatSpace(required: boolean = false) {
		const start = this.#offset;
		while (true) {
			const ch = this.#content[this.#offset];
			if (ch === ' ' || ch === '\t' || ch === '\n' || ch === '\r') this.go(1);
			else break;
		}
		if (required && start === this.#offset) {
			throw new ParseError(new Location(this.#line, this.#column, this.#offset), `此处应该有空白`);
		}
	}

	private eatString() {
		const start = new Location(this.#line, this.#column, this.#offset);
		const quote = this.#content[this.#offset];
		if (quote !== '"' && quote !== '\'') throw new ParseError(start, `字符串应该以双引号或单引号开始`);
		let finished = false;
		this.go(1);
		while (!this.eof()) {
			const ch = this.#content[this.#offset];
			if (ch === quote) {
				this.go(1);
				finished = true;
				break;
			}
			else if (ch === '\\') this.go(2);
			else this.go(1);
		}
		const end = new Location(this.#line, this.#column, this.#offset);
		if (!finished) throw new ParseError(end, `字符串应该以双引号或单引号结束`);
		return { start, end, text: this.#content.slice(start.offset, end.offset) };
	}

	private eatComment() {
		let finished = false;
		let start = this.#offset;
		let end = start;
		while (!this.eof()) {
			if (this.is('-->')) {
				end = this.#offset;
				this.go(3);
				finished = true;
				break;
			}
			this.go(1);
		}
		if (!finished) throw new ParseError(new Location(this.#line, this.#column, this.#offset), `注释应该以 --> 结束`);
		return { text: this.#content.slice(start, end) };
	}

	#isIDentifierChar(code: number) {
		if ((code >= 97 && code <= 122) || // a-z
			(code >= 65 && code <= 90) || // A-Z
			(code >= 48 && code <= 57) || // 0-9
			code === 95 || // _
			code === 45) // -
			return true;
		return false;
	}

	private eatIdentifier() {
		const start = new Location(this.#line, this.#column, this.#offset);
		while (true) {
			const code = this.#content.codePointAt(this.#offset);
			if (!code) break;
			if (!this.#isIDentifierChar(code)) break;
			this.go(1);
		}
		const end = new Location(this.#line, this.#column, this.#offset);
		if (start.offset === end.offset) throw new ParseError(start, `标识符不能为空`);
		return { start, end, text: this.#content.slice(start.offset, end.offset) };
	}

	private eatText() {
		const start = this.#offset;
		while (!this.eof() && !this.is('<')) this.go(1);
		const end = this.#offset;
		return { start, end, text: this.#content.slice(start, end) };
	}

	#parse(): Array<Element | Text | Comment> {
		const items: Array<Element | Text | Comment> = [];
		while (!this.eof()) {
			const start = new Location(this.#line, this.#column, this.#offset);
			// DOCTYPE
			if (this.eat('<!DOCTYPE')) {
				const items = this.#parseDoctype();
				this.#document.doctype = new DocType(new Position(start, new Location(this.#line, this.#column, this.#offset)), items);
			}
			// 注释
			else if (this.eat("<!--")) {
				const { text } = this.eatComment();
				items.push(new Comment(new Position(start, new Location(this.#line, this.#column, this.#offset)), text));
			}
			//结束
			else if (this.is("</")) break;
			// 标签
			else if (this.eat("<")) {
				const { tag, attrs, closed } = this.#parseElement();
				//闭合
				if (closed || ["br", "hr", "img", "input", "meta", "link", "base", "col", "command", "param", "source", "track", "wbr"].includes(tag.tagName.toLowerCase())) {
					items.push(new Element(
						new Position(start, new Location(this.#line, this.#column, this.#offset)),
						tag, attrs, true, [],
					))
				}
				//非闭合
				else {
					const children = this.#parse();
					if (!this.eat("</" + tag.tagName + ">")) throw new ParseError(new Location(this.#line, this.#column, this.#offset), `标签 ${tag.tagName} 应该有结束标签`);
					items.push(new Element(
						new Position(start, new Location(this.#line, this.#column, this.#offset)),
						tag, attrs, false, children,
					));
				}
			}
			//文本
			else {
				const { text } = this.eatText();
				items.push(new Text(new Position(start, new Location(this.#line, this.#column, this.#offset)), text));
			}
		}
		return items;
	}

	#parseDoctype() {
		this.eatSpace(true);
		const name = this.eatIdentifier();
		const items: string[] = [];
		while (!this.eof() && !this.eat('>')) {
			this.eatSpace(true);
			if (this.is('"') || this.is("'")) {
				const item = this.eatString();
				items.push(item.text);
			}
			else {
				const item = this.eatIdentifier();
				items.push(item.text);
			}
		}
		return [name.text, ...items];
	}

	#parseElement() {
		const name = this.eatIdentifier();
		const attrs: Array<ElementAttribute> = [];
		let closed = false;
		let gotBooleanAttribute = false;
		while (!this.eat('>')) {
			if (this.eat("/>")) {
				closed = true;
				break;
			}
			this.eatSpace(!gotBooleanAttribute);
			let isControlAttribute = false;
			if (this.eat(":")) isControlAttribute = true;
			else if (this.eat("/>")) {
				closed = true;
				break;
			}
			const name = this.eatIdentifier();
			this.eatSpace(false);
			if (this.eat("=")) {
				gotBooleanAttribute = false;
				this.eatSpace(false);
				const value = this.eatString();
				attrs.push(new ElementAttribute(
					new ElementAttributeName(new Position(name.start, name.end), name.text),
					new ElementAttributeValue(new Position(value.start, value.end), value.text),
					isControlAttribute,
				));
			}
			else {
				gotBooleanAttribute = true;
				attrs.push(new ElementAttribute(new ElementAttributeName(new Position(name.start, name.end), name.text), null, isControlAttribute));
			}
		}
		return {
			tag: new ElementTag(new Position(name.start, new Location(this.#line, this.#column, this.#offset)), name.text),
			attrs,
			closed,
		};
	}
}