import path from "path";
import { __system__, check } from "./util";

export abstract class HTMLNode {
	public abstract build(...args: string[]): string;
}

export class HTMLDocument extends HTMLNode {
	#doctype: string | null = null;
	#children: Array<HTMLElement | HTMLComment | HTMLText> = [];
	#templates: Array<HTMLTemplate> = [];

	public setDoctype(doctype: string) {
		this.#doctype = doctype;
		return this;
	}

	public addChild(child: HTMLElement | HTMLComment | HTMLText) {
		this.#children.push(child);
		return this;
	}

	public addTemplate(template: HTMLTemplate) {
		this.#templates.push(template);
		return this;
	}

	public get doctype() { return this.#doctype; }
	public get templates() { return this.#templates; }
	public get children() { return this.#children; }

	public build() {
		const buffer: string[] = [];
		if (this.doctype) buffer.push(this.doctype);
		for (const child of this.children) buffer.push(child.build());
		return buffer.join("");
	}
}

export class HTMLElement extends HTMLNode {
	#tag: string;
	#attributes: Array<{ name: string, value: string | null }> = [];
	#controls: Array<{ name: string, value: string }> = [];
	#children: Array<HTMLElement | HTMLComment | HTMLText> = [];
	#closed: boolean;
	#ifKey: number | null = null;

	constructor(tag: string, closed: boolean) {
		super();
		this.#tag = tag;
		this.#closed = closed;
	}

	public addAttribute(name: string, value: string | null) {
		this.#attributes.push({ name, value });
		return this;
	}

	public addControl(name: string, value: string) {
		if (!value) value = "false";
		if (value.startsWith("'") || value.startsWith('"')) value = value.slice(1, -1);
		this.#controls.push({ name, value });
		return this;
	}

	public setIf(key: number | null) {
		this.#ifKey = key;
		return this;
	}

	public get ifKey() { return this.#ifKey }

	public addChild(child: HTMLElement | HTMLComment | HTMLText) {
		this.#children.push(child);
		return this;
	}

	public get tag() { return this.#tag; }
	public get attributes() { return this.#attributes; }
	public get controls() { return this.#controls; }
	public get closed() { return this.#closed; }
	public get children() { return this.#children; }


	public build() {
		const resolveControl = (controls: Array<{ name: string, value: string }>, action: () => string): string => {
			if (!controls.length) return action();
			const [control, ...rest] = controls;

			if (control.name === "show") {
				return `\${__system__.if(${check(control.value, "false")}, null, __context__, ()=>\`${resolveControl(rest, action)}\`)}`;
			}
			if (control.name === "hide") {
				return `\${__system__.if(${check(`!(${control.value})`, "false")}, null, __context__, ()=>\`${resolveControl(rest, action)}\`)}`;
			}
			else if (control.name === "for") {
				const inIndex = control.value.indexOf(" in ");
				if (inIndex <= 0) return "";
				const itVal = control.value.slice(0, inIndex).trim();
				const itData = control.value.slice(inIndex + 4).trim().split(" as ");
				if (!itVal.length || !itData.length) return "";

				let itKey: string | null = null;
				let itValue: string | null = null;
				if (/^[a-zA-Z_][a-zA-Z0-9_]*$/.test(itVal)) itValue = itVal;
				else {
					const match = itVal.match(/^\(\s*([a-zA-Z0-9_][a-zA-Z0-9_]*)\s*,\s*([a-zA-Z0-9_][a-zA-Z0-9_]*)\s*\)$/);
					if (match) {
						itKey = match[1];
						itValue = match[2];
					}
				}
				if (!itValue) return "";
				return `\${__system__.for(${check(itData)}, (${itValue}${itKey ? `, ${itKey}` : ""})=>\`${resolveControl(rest, action)}\`)}`;
			}
			else if (control.name === "if") {
				return `\${__system__.if(${check(control.value)}, ${this.ifKey}, __context__, ()=>\`${resolveControl(rest, action)}\`)}`;
			}
			else if (control.name === "else") {
				if (!this.ifKey) return "";
				if (control.value.length) return `\${__system__.elseIf(${check(control.value)}, ${this.ifKey}, __context__, ()=>\`${resolveControl(rest, action)}\`)}`;
				else return `\${__system__.else(${this.ifKey}, __context__, ()=>\`${resolveControl(rest, action)}\`)}`;
			}
			else return `\`\``;
		}

		return resolveControl(this.controls, () => {
			if (this.tag === "use") {
				const from = this.attributes.find(a => a.name === "from")?.value;
				const name = this.attributes.find(a => a.name === "name")?.value;
				const data = this.attributes.find(a => a.name === "data")?.value;
				return `\${__system__.use(${from ? JSON.stringify(from) : "__filename__"}, ${JSON.stringify(name)}, ${data ?? "null"}, __dirname__)}`
			}
			else {
				const buffer = [`<${this.tag}`];
				for (const { name, value } of this.attributes) {
					if (value === null) buffer.push(` ${name}`);
					else buffer.push(` ${name}=${value.replace(/\{\{(.+?)\}\}/g, "${$1}")}`);
				}
				if (this.closed) buffer.push(` />`);
				else {
					buffer.push(">");
					for (const child of this.children) {
						buffer.push(child.build());
					}
					buffer.push(`</${this.tag}>`);
				}
				return buffer.join("");
			}
		});
	}
}

export class HTMLText extends HTMLNode {
	#text: string;
	constructor(text: string) {
		super();
		this.#text = text;
	}

	public get text() { return this.#text; }


	public build() {
		return this.text.replace(/\`/g, "\\`").replace(/\$/g, "\\$").replace(/\{\{(.+?)\}\}/g, "${$1}");
	}
}

export class HTMLComment extends HTMLNode {
	#comment: string;
	constructor(comment: string) {
		super();
		this.#comment = comment;
	}
	public get comment() { return this.#comment; }


	public build() {
		return `<!--${this.comment}-->`;
	}
}

export class HTMLTemplate extends HTMLNode {
	#name: string;
	#argument: string;
	#children: Array<HTMLElement | HTMLComment | HTMLText> = [];
	#filename: string;
	#cache: ((data: any) => string) | null = null;

	constructor(name: string, argument: string, filename: string) {
		super();
		this.#name = name;
		this.#argument = argument;
		this.#filename = path.resolve(process.cwd(), filename);
	}

	public addChild(child: HTMLElement | HTMLComment | HTMLText) {
		this.#children.push(child);
		return this;
	}

	public get name() { return this.#name; }
	public get argument() { return this.#argument; }
	public get children() { return this.#children; }

	public build() {
		return this.children.map(child => child.build()).join("");
	}

	public resolve() {
		if (!this.#cache) {
			const body = this.build();
			const func = new Function("data", "__system__", "__context__", "__filename__", "__dirname__", `return \`${body}\``);
			this.#cache = (data: any) => func(data, __system__, { if: 0 }, this.#filename, path.dirname(this.#filename));
		}
		return this.#cache;
	}
}