
import path from "path";
import { Comment, Element, Parser, Text } from "../parser";
import { HTMLComment, HTMLDocument, HTMLElement, HTMLTemplate, HTMLText } from "./element";
import { __system__ } from "./util";
import { renderConfig } from "./config";

export class Module {
	#parser: Parser;
	#document: HTMLDocument;
	#cache: ((data: any) => string) | null = null;
	static #modules: Map<string, Module> = new Map();

	public static get(filename: string) {
		if (path.extname(filename) !== renderConfig.extname) filename += renderConfig.extname;
		//使用缓存
		if (renderConfig.cache) {
			if (!this.#modules.has(filename)) {
				this.#modules.set(filename, new Module(filename));
			}
			return Module.#modules.get(filename)!;
		}
		//不是有缓存
		else return new Module(filename);
	}

	private constructor(public readonly filename: string) {
		this.#parser = Parser.fromFile(filename);
		this.#document = new HTMLDocument();
		const doc = this.#parser.parse();
		if (doc.doctype) this.#document.setDoctype(`<!DOCTYPE ${doc.doctype.items.join(' ')}>`);
		this.#resolve(this.#document, this.#document, doc.children, { if: 0 });
		Module.#modules.set(filename, this);
	}

	public template(name: string) {
		return this.#document.templates.find(t => t.name === name) ?? null;
	}

	#findLastElement(children: Array<HTMLElement | HTMLComment | HTMLText>): HTMLElement | null {
		for (let i = children.length - 1; i >= 0; i--) {
			if (children[i] instanceof HTMLElement) return children[i] as HTMLElement;
		}
		return null;
	}

	#resolve(root: HTMLDocument, parent: HTMLElement | HTMLDocument | HTMLTemplate, children: Array<Element | Text | Comment>, context: { if: number }) {
		for (const child of children) {
			if (child instanceof Element) {
				const tag = child.tagName;
				const lowerCaseTag = tag.toLowerCase();

				// 模板
				if (lowerCaseTag === "template") {
					const nameAttr = child.attributes.find(a => a.name.name === "name");
					if (!nameAttr?.value) throw new Error("template element must have a name attribute");
					let name = nameAttr.value.value;
					if (name.startsWith("'") || name.startsWith('"')) name = name.slice(1, -1);

					const argumentAttr = child.attributes.find(a => a.name.name === "argument");
					let argument = "data";
					if (argumentAttr?.value) {
						const value = argumentAttr.value.value;
						if (value.startsWith("'") || value.startsWith('"')) argument = value.slice(1, -1);
						else argument = value;
					}
					const template = new HTMLTemplate(name, argument, this.filename);
					root.addTemplate(template);

					if (child.children.length) this.#resolve(root, template, child.children, context);
				}

				// 模板引用
				else if (lowerCaseTag === "use") {
					const fromAttr = child.attributes.find(a => a.name.name === "from");
					let from: string | null = null;
					if (fromAttr) {
						if (!fromAttr.value) throw new Error("use element must have a from attribute");
						from = fromAttr.value.value;
						if (from.startsWith("'") || from.startsWith('"')) from = from.slice(1, -1);
					}

					const nameAttr = child.attributes.find(a => a.name.name === "name");
					if (!nameAttr?.value) throw new Error("use element must have a name attribute");
					let name = nameAttr.value.value;
					if (name.startsWith("'") || name.startsWith('"')) name = name.slice(1, -1);

					const dataAttr = child.attributes.find(a => a.name.name === "data");
					let data: string | null = null;
					if (dataAttr?.value) {
						const value = dataAttr.value.value;
						data = value.startsWith("'") || value.startsWith('"') ? value.slice(1, -1) : value;
					}

					const use = new HTMLElement("use", true);
					if (from) use.addAttribute("from", from);
					use.addAttribute("name", name);
					if (data) use.addAttribute("data", data);

					for (const attr of child.attributes) {
						const name = attr.name.name;
						if (attr.control) {
							use.addControl(name, attr.value?.value ?? "");
							if (name === "if") use.setIf(++context.if);
							else if (name === "else") use.setIf(this.#findLastElement(parent.children)?.ifKey ?? null);
						}
					}

					parent.addChild(use);
				}

				//普通元素
				else {
					const element = new HTMLElement(tag, child.selfClosing);
					for (const attr of child.attributes) {
						const name = attr.name.name;
						if (attr.control) {
							element.addControl(name, attr.value?.value ?? "");
							if (name === "if") element.setIf(++context.if);
							else if (name === "else") element.setIf(this.#findLastElement(parent.children)?.ifKey ?? null);
						}
						else element.addAttribute(name, attr.value?.value ?? null);
					}
					parent.addChild(element);
					this.#resolve(root, element, child.children, context);
				}
			}
			//文本节点
			else if (child instanceof Text) {
				parent.addChild(new HTMLText(child.value));
			}
			//注释
			else if (child instanceof Comment) {
				parent.addChild(new HTMLComment(child.value));
			}
		}
	}

	resolve() {
		if (!this.#cache) {
			const body = this.#document.build();
			const func = new Function("data", "__system__", "__context__", "__filename__", "__dirname__", `return \`${body}\`;`);
			this.#cache = (data: any) => func(data, __system__, { if: 0 }, this.filename, path.dirname(this.filename));
		}
		return this.#cache;
	}
}
